/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2012, Diaple Networking SL
 *
 * Roberto Casas <roberto.casas@diaple.com>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*! \file
 *
 * \brief Wait until a signal is received
 *
 * \author Roberto Casas Millan <roberto.casas@diaple.com>
 *
 * \ingroup applications
 */

#include "asterisk.h"

ASTERISK_FILE_VERSION(__FILE__, "$Revision:  $")

#include "asterisk/logger.h"
#include "asterisk/channel.h"
#include "asterisk/pbx.h"
#include "asterisk/module.h"
#include "asterisk/lock.h"
#include "asterisk/app.h"


/*** DOCUMENTATION
        <application name="WaitCond" language="en_US">
                <synopsis>
                        Wait until a signal is received. If a timeout is specified, exit if reached.
                </synopsis>
                <syntax>
                        <parameter name="condname" required="true" />
                        <parameter name="timeout" required="false" />
                </syntax>
                <description>
                        <para>Wait until the a signal is received or the timeout (if specified) is reached.</para>
                        <para>Sets <variable>WAITCONDSTATUS</variable> to one of the following values:</para>
                        <variablelist>
                                <variable name="WAITCONDSTATUS">
                                        <value name="OK">
                                                Wait succeeded (signal received).
                                        </value>
                                        <value name="FAILURE">
                                                Invalid argument.
                                        </value>
                                        <value name="PAST">
                                                Time specified had already past.
                                        </value>
                                </variable>
                        </variablelist>
                </description>
        </application>
        <application name="CondSignal" language="en_US">
                <synopsis>
                        Send a signal to wake up threads waiting in a WaitCond.
                </synopsis>
                <syntax>
                        <parameter name="condname" required="true" />
                </syntax>
                <description>
                        <para>Send a signal to wake up threads waiting in a WaitCond.</para>
                </description>
        </application>
 ***/

static char *wait_app = "WaitCond";
static char *signal_app = "CondSignal";

static int unloading = 0;

struct waitcond_frame {
    AST_LIST_ENTRY(waitcond_frame) entries;
    ast_mutex_t mutex;
    ast_cond_t cond;
    char name[0];
};

static AST_LIST_HEAD_STATIC(conditions, waitcond_frame);

static int waitcond_exec(struct ast_channel *chan, const char *data) {
    AST_DECLARE_APP_ARGS(args,
            AST_APP_ARG(condname);
            AST_APP_ARG(waittime);
            );

    int res;
    char *parse;

    parse = ast_strdupa(data);
    AST_STANDARD_APP_ARGS(args, parse);


    if (ast_strlen_zero(args.condname)) {
        ast_log(LOG_WARNING, "WaitCond requires an argument(condname)\n");
        pbx_builtin_setvar_helper(chan, "WAITCONDSTATUS", "FAILURE");
        return 0;
    }

    int ms = 0;

    if (!ast_strlen_zero(args.waittime)) {
        if (ast_app_parse_timelen(args.waittime, &ms, TIMELEN_SECONDS) != 0) {
            ast_log(LOG_WARNING, "WaitCond timeout argument must be numeric\n");
            pbx_builtin_setvar_helper(chan, "WAITCONDSTATUS", "FAILURE");
            return 0;
        }
    }

    struct waitcond_frame *current;

    /* Lock already exists? */
    AST_LIST_LOCK(&conditions);

    AST_LIST_TRAVERSE(&conditions, current, entries) {
        if (strcmp(current->name, args.condname) == 0) {
            break;
        }
    }

    if (!current) {
        if (unloading) {
            /* Don't bother */
            AST_LIST_UNLOCK(&conditions);
            return -1;
        }

        /* Create new lock entry */
        current = ast_calloc(1, sizeof (*current) + strlen(args.condname) + 1);
        if (!current) {
            AST_LIST_UNLOCK(&conditions);
            return -1;
        }

        strcpy(current->name, args.condname); /* SAFE */
        if ((res = ast_mutex_init(&current->mutex))) {
            ast_log(LOG_ERROR, "Unable to initialize mutex: %s\n", strerror(res));
            ast_free(current);
            AST_LIST_UNLOCK(&conditions);
            return -1;
        }
        if ((res = ast_cond_init(&current->cond, NULL))) {
            ast_log(LOG_ERROR, "Unable to initialize condition variable: %s\n", strerror(res));
            ast_mutex_destroy(&current->mutex);
            ast_free(current);
            AST_LIST_UNLOCK(&conditions);
            return -1;
        }
        AST_LIST_INSERT_TAIL(&conditions, current, entries);
    }
    AST_LIST_UNLOCK(&conditions);

    AST_LIST_LOCK(&conditions);
    ast_mutex_lock(&current->mutex);
    AST_LIST_UNLOCK(&conditions);

    if (ms > 0) {
        struct timespec timeout;
        struct timeval now;

        now = ast_tvadd(ast_tvnow(), ast_samp2tv(ms, 1000));
        timeout.tv_sec = now.tv_sec;
        timeout.tv_nsec = now.tv_usec * 1000;

        res = ast_cond_timedwait(&current->cond, &current->mutex, &timeout);
    } else {
        res = ast_cond_wait(&current->cond, &current->mutex);
    }
    ast_mutex_unlock(&current->mutex);

    if (res == ETIMEDOUT) {
        pbx_builtin_setvar_helper(chan, "WAITCONDSTATUS", "TIMEDOUT");
    } else if (res == 0) {
        pbx_builtin_setvar_helper(chan, "WAITCONDSTATUS", "OK");
    }


    return res;
}

static int condsignal_exec(struct ast_channel *chan, const char *data) {
    AST_DECLARE_APP_ARGS(args,
            AST_APP_ARG(condname);
            );

    int res;

    char *parse;

    parse = ast_strdupa(data);
    AST_STANDARD_APP_ARGS(args, parse);

    struct waitcond_frame *current;

    /* Lock already exists? */
    AST_LIST_LOCK(&conditions);

    AST_LIST_TRAVERSE_SAFE_BEGIN(&conditions, current, entries) {
        if (strcmp(current->name, args.condname) == 0) {
            res = ast_cond_broadcast(&current->cond);
            AST_LIST_REMOVE_CURRENT(entries);
            break;
        }
    }
    AST_LIST_TRAVERSE_SAFE_END;

    AST_LIST_UNLOCK(&conditions);

    return res;
}

static int unload_module(void) {
    int res;

    res = ast_unregister_application(wait_app);
    res |= ast_unregister_application(signal_app);

    return res;
}

static int load_module(void) {
    int res;

    res = ast_register_application_xml(wait_app, waitcond_exec);
    res |= ast_register_application_xml(signal_app, condsignal_exec);

    return res;
}

AST_MODULE_INFO_STANDARD(ASTERISK_GPL_KEY, "Wait for signal");